function [Fi, CVs, Penalty] = NLPcalc(yji, FiMax, Bias, F, xGuess, Pmin, Pmax)
% Calculate optimal gas blending steady state by NLP
% Decision variables are:
% 1-6  Feed stream rates, kNm^3/h
%   7  Slack variable for constraint softening

Opts = optimset('fmincon');
Opts = optimset(Opts, 'Display', 'none', 'Algorithm', 'sqp', ...
    'GradObj', 'on');

% Default quality bounds are true process values.
if nargin < 6
    Pmin = [16.5; 25; 39];
end
if nargin < 7
    Pmax = [18.0; 27; 46];
end

% Current bias vector.  If bias not supplied for FSI, assume zero
if length(Bias) == 2
    bnow = [Bias(:); 0];
else
    bnow = Bias(:);
end
Param = struct('yji', {yji}, 'Bias', {bnow}, 'Pmin', {Pmin}, 'Pmax', {Pmax});
% Bounds on decision variables
LB = zeros(9,1);
UB = [FiMax(:); Inf; Inf; Inf];
% Equality constraint forces the feed rates to meet the demand.
Aeq = [ones(1,6) zeros(1,3)];
beq = F;
% Calculate the optimal feed ratios
x = fmincon(@BlendObjective, xGuess, [], [], Aeq, beq, LB, UB, ...
            @(x) BlendNLC(x, Param), Opts);
Fi = x(1:6);
Eps = x(7:9);
Penalty = 0.02*sum((Eps./(0.07*ones(3,1))).^2);
% Calculate the corresponding properties.
[~, ~, CVs] = BlendNLC(x, Param);

function [Obj, f] = BlendObjective(x)
% Objective function for gas blending steady-state NLP.  Objective is linear.
% Gradient is also returned.
% Relative costs
f = [0.678; 0.254; 0; 0.068; 0; 0];
Penalty = 0.02*sum((x(7:9)./(0.07*ones(3,1))).^2);
Obj = f'*x(1:6) + Penalty;
f = [f; 0.04*x(7:9)/0.0049];

function [c, ceq, Props] = BlendNLC(x, Param)
% Nonlinear constraints for gas blending NLP.  Feasible if c <= 0.

Fi = x(1:6);
Slack = x(7:9);   % Constraint softening to ensure a feasible solution
Ratios = Fi(:)/sum(Fi);
% Calculate steady state header composition for given feed rates and compositions
y = Param.yji*Ratios;
% Corresponding gas properties.  T, V and Z don't matter here.
Props = FuelGasProperties(y, 300, 100, 1.06);
% Account for apparant bias
Props = Props(2:4);
Props = Props(:) - Param.Bias(:);
c = [(Param.Pmin(:) - Slack) - Props; ...
    Props - (Param.Pmax(:) + Slack)];
ceq = [];   % No equality constraints


function [Props, dPropdN] = FuelGasProperties(Nmol, T, V, Z)

% Compute HHV, WI, and FSI in terms of kmol of each component in
% mixture.  Required sequence is C1, C2+, H2, N2, CO, CO2.

% Optional calculation of partial molal values

NT = sum(Nmol);
y = Nmol(:)/NT;
HHVj = [37.78, 126.5, 12.10, 0.00, 11.97, 0.00];
SGj = [16.04, 58.12, 2.016, 28.02, 28.01, 44.01]/28.8;
FNj = [148, 514, 339, 0, 61, 0];
FDj = [9.55, 31, 2.39, 5, 2.39, 5];
HHV = HHVj*y;
SG = SGj*y;
WI = HHV/sqrt(SG);
FN = FNj*y;
FD = FDj*y + 1;
FSI = FN/FD;
P = NT*8.314*T*Z/V;  % Pressure in kPa
Props = [P, HHV, WI, FSI];

if nargout > 1
    % Calculate partial molal properties
    dPdNk = 8.314*T*Z/V*ones(6,1);
    dHHVdNk = HHVj'/NT - ones(6,1)*HHV/NT;
    dSGdNk = SGj'/NT - ones(6,1)*SG/NT;
    dWIdNk = dHHVdNk/sqrt(SG) - 0.5*dSGdNk*WI/SG;
    dFNdNk = FNj'/NT - ones(6,1)*FN/NT;
    dFDdNk = FDj'/NT - ones(6,1)*(FD-1)/NT;
    dFSIdNk = (dFNdNk - dFDdNk*FSI)/FD;
    dPropdN = [dPdNk, dHHVdNk, dWIdNk, dFSIdNk];
end